/*
 * mks-dmabuf-paintable-private.h
 *
 * Copyright 2023 Christian Hergert <chergert@redhat.com>
 * Copyright 2023 Bilal Elmoussaoui <belmouss@redhat.com>
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of the
 * License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <gdk/gdk.h>

G_BEGIN_DECLS

typedef struct _MksDmabufScanoutData
{
  guint                  width;
  guint                  height;
  guint                  stride;
  guint                  fourcc;
  guint64                modifier;
  int                    dmabuf_fd;
} MksDmabufScanoutData;

#define MKS_TYPE_DMABUF_PAINTABLE (mks_dmabuf_paintable_get_type())

G_DECLARE_FINAL_TYPE (MksDmabufPaintable, mks_dmabuf_paintable, MKS, DMABUF_PAINTABLE, GObject)

MksDmabufPaintable *mks_dmabuf_paintable_new        (void);
gboolean            mks_dmabuf_paintable_import     (MksDmabufPaintable   *self,
                                                     GdkGLContext         *gl_context,
                                                     MksDmabufScanoutData *data,
                                                     cairo_region_t       *region,
                                                     GError              **error);

G_END_DECLS
