baseURLs = [
    [ 'GLib', 'https://docs.gtk.org/glib/' ],
    [ 'GObject', 'https://docs.gtk.org/gobject/' ],
    [ 'Gio', 'https://docs.gtk.org/gio/' ],
    [ 'Gdk', 'https://docs.gtk.org/gdk4/' ],
    [ 'Gsk', 'https://docs.gtk.org/gsk4/' ],
    [ 'Gtk', 'https://docs.gtk.org/gtk4/' ],
]
